/*
    Copyright 2005,2006 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "md5.h"

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define stricmp     strcasecmp
    #define stristr     strcasestr
    #define strnicmp    strncasecmp
#endif



#define VER         "0.1"
#define BUFFSZ      4096
#define HOST        "gpcm.gamespy.com"
#define PORT        29900

#define LOGIN       "\\login\\"             \
                    "\\challenge\\%s"       \
                    "\\uniquenick\\%s"      \
                    "\\response\\%s"        \
                    "\\port\\%hu"           \
                    "\\namespaceid\\1"      \
                    "\\id\\1"               \
                    "\\final\\"
#define LOGOUT      "\\logout\\"            \
                    "\\sesskey\\%d"         \
                    "\\final\\"
#define SPACES48    "                                                "
#define LOGINRESP   "%s"    /* MD5 password */      \
                    "%s"    /* 48 spaces */         \
                    "%s"    /* user nickname */     \
                    "%s"    /* client challenge */  \
                    "%s"    /* server challenge */  \
                    "%s"    /* MD5 password */      \



int gsrecv(int sd, u_char *data, int size);
u_char *gsparval(u_char *data, int size, u_char *par);
u_char *gs_response(u_char *nickname, u_char *password, u_char *client_challenge, u_char *server_challenge);
void do_md5(char *data, int len, char *out);
u_char *create_rand_string(u_int seed, u_char *data, int len);
u_int resolv(char *host);
void std_err(void);



int main(int argc, char *argv[]) {
    struct  sockaddr_in peer;
    u_int   seed;
    int     sd,
            len,
            sesskey;
    u_char  buff[BUFFSZ + 1],
            *nick,
            *pass,
            *response,
            *errmsg,
            client_chall[33],
            *server_chall;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stdout, NULL);
    setbuf(stderr, NULL);

    fputs("\n"
        "GS login checker "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 3) {
        printf("\n"
            "Usage: %s <username> <password>\n"
            "\n"
            "For any other player-related information on Gamespy use my gsplayers tool\n"
            "\n", argv[0]);
        exit(1);
    }

    nick = argv[1];
    pass = argv[2];

    printf("- resolve host %s\n", HOST);

    peer.sin_addr.s_addr = resolv(HOST);
    peer.sin_port        = htons(PORT);
    peer.sin_family      = AF_INET;

    printf("- %s : %hu\n",
        inet_ntoa(peer.sin_addr), ntohs(peer.sin_port));

    sd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
    if(sd < 0) std_err();
    printf("- connecting... ");
    if(connect(sd, (struct sockaddr *)&peer, sizeof(peer))
      < 0) std_err();
    printf("ok\n");

    printf("- receive challenge:\n");
    len = gsrecv(sd, buff, BUFFSZ);
    printf("  %s\n", buff);

    server_chall = gsparval(buff, len, "challenge");
    if(!server_chall) {
        printf("\nError: the server has not sent the challenge field\n\n");
        exit(1);
    }

    seed = time(NULL);

    create_rand_string(seed, client_chall, 32);

    response = gs_response(nick, pass, client_chall, server_chall);

    len = snprintf(buff, BUFFSZ, LOGIN,
        client_chall,
        nick,
        response,
        seed & 0x7fff);
    if((len < 0) && (len > BUFFSZ)) exit(1);

    printf("- send query:\n  %s\n", buff);
    if(send(sd, buff, len, 0)
      < 0) std_err();

    printf("- receive answer:\n");
    len = gsrecv(sd, buff, BUFFSZ);
    printf("  %s\n", buff);

    if(gsparval(buff, len, "error")) {  // error
        errmsg = gsparval(buff, len, "errmsg");
        printf("- %s\n", errmsg ? (char *)errmsg : "server has not accepted your query, invalid account");
        free(errmsg);
        printf("\n  Your account is NOT valid\n\n");
    } else {
        errmsg = gsparval(buff, len, "sesskey");
        if(errmsg) {
            sesskey = atoi(errmsg);
            free(errmsg);
            len = sprintf(buff, LOGOUT, sesskey);
            printf("- send logout:\n  %s\n", buff);
            if(send(sd, buff, len, 0)
              < 0) std_err();
        }
        printf("\n  Your account IS valid!!!\n\n");
    }

    close(sd);
    return(0);
}



int gsrecv(int sd, u_char *data, int size) {
    int     t,
            len;
    u_char  *p,
            *s,
            *val;

    p = data;
    s = data;
    len = 0;

    while(size) {
        t = recv(sd, p, 1, 0);
        if(t < 0) std_err();
        if(!t) break;

        len++;  p++;    size--;

        if(*(p - 1) == '\\') {
            val = gsparval(data, len, "final");
            if(val) {
                free(val);
                break;
            }
        }
    }
    *p = 0;
    return(len);
}



u_char *gsparval(u_char *data, int size, u_char *par) {
    int     nt    = 1,
            found = 0,
            len,
            parlen;
    u_char  *p,
            *limit,
            *val  = NULL;

    if(size < 0) size = strlen(data);   // -1 for NULL limited strings
    limit = data + size;
    parlen = strlen(par);

    for(p = data; data < limit; data = p + 1, nt++) {
        for(p = data; (p < limit) && (*p != '\\'); p++);
        len = p - data;

        if(nt & 1) {    // val
            if(found) {
                val = malloc(len + 1);
                memcpy(val, data, len);
                val[len] = 0;
                found = 0;
            }
        } else {        // par
            if(!strnicmp(data, par, parlen)) found = 1;
        }
    }

    if(found && !val) val = strdup("");
    return(val);
}



u_char *gs_response(u_char *nickname, u_char *password, u_char *client_challenge, u_char *server_challenge) {
    int     len;
    u_char  pwdmd5[33],
            buff[BUFFSZ + 1];
    static u_char   retmd5[33];

    do_md5(password, strlen(password), pwdmd5);

    len = snprintf(buff, BUFFSZ, LOGINRESP,
        pwdmd5,
        SPACES48,
        nickname,
        client_challenge,
        server_challenge,
        pwdmd5);
    if((len < 0) && (len > BUFFSZ)) exit(1);

    do_md5(buff, len, retmd5);
    return(retmd5);
}



void do_md5(char *data, int len, char *out) {
    md5_context md5t;
    int     i,
            x;
    u_char  md5h[16];
    static const char  hex[16] = "0123456789abcdef";

    md5_starts(&md5t);
    md5_update(&md5t, data, len);
    md5_finish(&md5t, md5h);
    for(i = 0, x = 0; i < 16; i++) {
        out[x++] = hex[md5h[i] >> 4];
        out[x++] = hex[md5h[i] & 0xf];
    }
    out[x] = 0;
}



u_char *create_rand_string(u_int seed, u_char *data, int len) {
    u_char  *p;
    static const u_char table[] =
        "0123456789"
        "abcdefghijklmnopqrstuvwxyz"
        "ABCDEFGHIJKLMNOPQRSTUVWXYZ";

    for(p = data; len; p++, len--) {
        seed = (seed * 0x343FD) + 0x269EC3;
        *p = table[seed % (sizeof(table) - 1)];
    }
    *p = 0;
    return(data);
}



u_int resolv(char *host) {
    struct  hostent *hp;
    u_int   host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            printf("\nError: Unable to resolve hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u_int *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif

